from __future__ import unicode_literals
from __future__ import print_function
import time

from pytest import mark, fixture

from concrete.inspect import (
    _reconcile_index_and_tool,
    _valid_index_lun,
    _get_conll_deprel_tags_for_tokenization as cdt,
    print_situation_mentions,
    print_conll_style_tags_for_communication,
)
from concrete.util import generate_UUID as genId, create_comm
from concrete.util.references import add_references_to_communication
from concrete import (
    AnnotationMetadata,
    Communication,
    Dependency,
    DependencyParse,
    EntityMention,
    EntityMentionSet,
    MentionArgument,
    Property,
    Section,
    Sentence,
    SituationMention,
    SituationMentionSet,
    TextSpan,
    TokenRefSequence,
    Tokenization,
    TokenizationKind,
    Token,
    TokenList,
    TokenTagging,
    TaggedToken,
    UUID,
)


def _comm_with_properties(num_properties):
    ts = 17
    meta_tokn = AnnotationMetadata(tool='tokn-tool', timestamp=ts)
    toks = TokenList(tokenList=[Token(tokenIndex=0,
                                      text='text',
                                      textSpan=TextSpan(start=0,
                                                        ending=1))])
    tokn = Tokenization(uuid=genId(), metadata=meta_tokn,
                        kind=TokenizationKind.TOKEN_LIST,
                        tokenList=toks)
    sentence = Sentence(uuid=genId(), tokenization=tokn)
    section = Section(uuid=genId(), kind='kind', label='label',
                      sentenceList=[sentence])
    trfs = TokenRefSequence(tokenizationId=tokn.uuid,
                            tokenIndexList=[0],
                            anchorTokenIndex=0)
    em = EntityMention(uuid=genId(), entityType='entityType',
                       text='text', tokens=trfs)
    meta_ems = AnnotationMetadata(tool='ems-tool', timestamp=ts)
    ems = EntityMentionSet(uuid=genId(), metadata=meta_ems,
                           mentionList=[em])
    meta_prop = AnnotationMetadata(tool='Annotator1',
                                   timestamp=ts)
    props = list(
        Property(
            value="Property%d" % i,
            metadata=meta_prop,
            polarity=4.0) for i in range(num_properties))
    am = MentionArgument(role='role', entityMentionId=em.uuid,
                         propertyList=props)
    sm = SituationMention(uuid=genId(), text='text',
                          situationType='stiuationType',
                          situationKind='situationKind',
                          tokens=trfs, argumentList=[am])
    meta_sms = AnnotationMetadata(tool='sms-tool', timestamp=ts)
    sms = SituationMentionSet(uuid=genId(), metadata=meta_sms,
                              mentionList=[sm])
    meta_comm = AnnotationMetadata(tool='tool', timestamp=ts)
    comm = Communication(uuid=genId(), id='id', text='text',
                         type='type', metadata=meta_comm,
                         sectionList=[section],
                         situationMentionSetList=[sms],
                         entityMentionSetList=[ems])
    add_references_to_communication(comm)
    return comm


@fixture
def comm_with_other_tags():
    comm = create_comm('quick', '''\
The quick brown fox jumped
over the lazy dog .

Or did she ?
''')
    for section in comm.sectionList:
        for sentence in section.sentenceList:
            sentence.tokenization.tokenTaggingList = [
                TokenTagging(
                    uuid=genId(),
                    metadata=AnnotationMetadata(
                        tool=u'tool',
                        timestamp=1,
                    ),
                    taggingType=u'upper',
                    taggedTokenList=[
                        TaggedToken(
                            tokenIndex=token.tokenIndex,
                            tag=token.text.upper(),
                        )
                        for token in sentence.tokenization.tokenList.tokenList
                    ],
                ),
                TokenTagging(
                    uuid=genId(),
                    metadata=AnnotationMetadata(
                        tool=u'tool',
                        timestamp=1,
                    ),
                    taggingType=u'lower',
                    taggedTokenList=[
                        TaggedToken(
                            tokenIndex=token.tokenIndex,
                            tag=token.text.lower(),
                        )
                        for token in sentence.tokenization.tokenList.tokenList
                    ],
                ),
            ]
    return comm


def test_print_conll_other_tags_ignore_all(capsys, comm_with_other_tags):
    print_conll_style_tags_for_communication(comm_with_other_tags)
    (out, err) = capsys.readouterr()
    assert err == ''
    assert out.startswith(
        'INDEX\tTOKEN\n'
        '-----\t-----\n'
        '1\tThe\n'
        '2\tquick\n'
    )


def test_print_conll_other_tags_ignore_some(capsys, comm_with_other_tags):
    print_conll_style_tags_for_communication(comm_with_other_tags,
                                             other_tags=('upper',))
    (out, err) = capsys.readouterr()
    assert err == ''
    assert out.startswith(
        'INDEX\tTOKEN\tupper\n'
        '-----\t-----\t-----\n'
        '1\tThe\tTHE\n'
        '2\tquick\tQUICK\n'
    )
    assert '3\tshe\tSHE\n' in out


def test_print_conll_other_tags(capsys, comm_with_other_tags):
    print_conll_style_tags_for_communication(comm_with_other_tags,
                                             other_tags=('upper', 'lower'))
    (out, err) = capsys.readouterr()
    assert err == ''
    assert out.startswith(
        'INDEX\tTOKEN\tupper\tlower\n'
        '-----\t-----\t-----\t-----\n'
        '1\tThe\tTHE\tthe\n'
        '2\tquick\tQUICK\tquick\n'
    )
    assert '3\tshe\tSHE\tshe\n' in out


class ReconcileIndexAndTool:

    default_tool = "My awesome tool"

    def __init__(self, num, t=default_tool):
        self.lst = [
            DependencyParse(
                uuid=UUID(uuidString=str(1)),
                metadata=AnnotationMetadata(
                    timestamp=int(time.time()),
                    tool=t if i == 0 else "tool_" + str(i)
                ),
                dependencyList=[]
            ) for i in range(num)
        ] if num is not None else None


def test_print_situation_mentions_without_properties(capsys):
    print_situation_mentions(_comm_with_properties(0))
    (out, err) = capsys.readouterr()
    assert 'Properties' not in out


@mark.parametrize('num_properties', [(1,), (2,), (3,)])
def test_print_situation_mentions_with_properties(capsys, num_properties):
    num_properties = 3
    print_situation_mentions(_comm_with_properties(num_properties))
    (out, err) = capsys.readouterr()
    assert 1 == out.count("Properties")
    assert 1 == out.count("Annotator1")
    assert num_properties == out.count("Property")
    assert num_properties == out.count("4.0")


def test_reconcile_index_and_tool_none_list():
    lst = ReconcileIndexAndTool(None).lst
    assert _reconcile_index_and_tool(lst, 0, None) == -1
    assert _reconcile_index_and_tool(lst, -2, None) == -1
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == -1


def test_reconcile_index_and_tool_empty_list():
    lst = ReconcileIndexAndTool(0).lst
    assert _reconcile_index_and_tool(lst, 0, None) == -1
    assert _reconcile_index_and_tool(lst, -2, None) == -1
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == -1


def test_reconcile_index_and_tool_singleton_list():
    lst = ReconcileIndexAndTool(1).lst
    assert _reconcile_index_and_tool(lst, 0, None) == 0
    assert _reconcile_index_and_tool(lst, -2, None) == -2
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == 0
    assert _reconcile_index_and_tool(lst, 0, "Your awesome tool") == -1


def test_reconcile_index_and_tool_three_list():
    lst = ReconcileIndexAndTool(3).lst
    assert _reconcile_index_and_tool(lst, 0, None) == 0
    assert _reconcile_index_and_tool(lst, -2, None) == -2
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == 0
    assert _reconcile_index_and_tool(lst, 0, "Your awesome tool") == -1
    tmp = lst[0]
    lst[0] = lst[2]
    lst[2] = tmp
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == 2
    assert _reconcile_index_and_tool(lst, 3, "My awesome tool") == 2
    assert _reconcile_index_and_tool(lst, 3, "Your awesome tool") == -1


def test_valid_index_lun_none_list():
    assert not _valid_index_lun(None, 0)
    assert not _valid_index_lun(None, -2)
    assert not _valid_index_lun(None, 0)


def test_valid_index_lun_empty_list():
    assert not _valid_index_lun([], 0)
    assert not _valid_index_lun([], -2)
    assert not _valid_index_lun([], 0)


def test_valid_index_lun_singleton_list():
    assert _valid_index_lun(range(1), 0)
    assert not _valid_index_lun(range(1), -2)
    assert not _valid_index_lun(range(1), 1)


class FakeTokenizationAndDependency:

    default_tool = "My awesome tool"

    def __init__(self, num, num_dp=1, t=default_tool):
        self.tokenization = Tokenization(
            uuid=UUID(uuidString=u"id"),
            metadata=AnnotationMetadata(
                timestamp=int(time.time()),
                tool=u"tool"
            ),
            kind=TokenizationKind.TOKEN_LIST,
        )
        if num is not None:
            self.tokenization.tokenList = TokenList(
                tokenList=[Token(tokenIndex=i,
                                 text=str(i))
                           for i in range(num)]
            )
            self.tokenization.dependencyParseList = [
                DependencyParse(
                    uuid=UUID(uuidString=u"depparse_" + str(num)),
                    metadata=AnnotationMetadata(
                        timestamp=int(time.time()),
                        tool=t if dp_idx == 0 else
                        ("dptool" + str(dp_idx))
                    ),
                    dependencyList=[
                        Dependency(
                            gov=i-1, dep=i,
                            edgeType="edge_" + str(i) +
                            "/" + str(dp_idx)
                        ) for i in range(num)
                    ]
                ) for dp_idx in range(num_dp)
            ]


def test_get_conll_deprel_tags_none_tokens():
    tokenization = FakeTokenizationAndDependency(None).tokenization
    for i in range(-1, 2):
        for tool in (None, "", FakeTokenizationAndDependency.default_tool):
            assert list(cdt(tokenization, i, tool)) == []


def test_get_conll_deprel_tags_zero_tokens():
    tokenization = FakeTokenizationAndDependency(0).tokenization
    for i in range(-1, 2):
        for tool in (None, "", FakeTokenizationAndDependency.default_tool):
            assert list(cdt(tokenization, i, tool)) == []


def test_get_conll_deprel_tags_one_tokens_found():
    tokenization = FakeTokenizationAndDependency(1).tokenization
    assert list(cdt(tokenization, 0, None)) == [u"edge_0/0"]
    assert (
        list(cdt(tokenization, 0, FakeTokenizationAndDependency.default_tool)) ==
        [u"edge_0/0"])
    assert list(cdt(tokenization, 0, "")) == [""]

    for tool in (None, ""):
        assert list(cdt(tokenization, -1, tool)) == [""]

    assert (
        list(cdt(tokenization, -1, FakeTokenizationAndDependency.default_tool)) ==
        [u"edge_0/0"])

    assert (
        list(cdt(tokenization, 1, FakeTokenizationAndDependency.default_tool)) ==
        [u"edge_0/0"])
    for tool in (None, ""):
        assert list(cdt(tokenization, 1, tool)) == [""]


def test_get_conll_deprel_tags_arbitrary_num_tokens_not_found():
    t = "Other great tool"
    for num_tokens in range(1, 10):
        tokenization = FakeTokenizationAndDependency(num_tokens,
                                                     t=t).tokenization
        assert list(cdt(tokenization, 0, None)) == ["edge_" + str(i) + "/0"
                                                    for i in range(num_tokens)]
        assert (
            list(cdt(tokenization, 0, FakeTokenizationAndDependency.default_tool)) ==
            [""] * num_tokens
        )
        assert list(cdt(tokenization, 0, "")) == [""] * num_tokens

        assert list(cdt(tokenization, -1, None)) == [""] * num_tokens
        assert tokenization.tokenList

        assert list(cdt(tokenization, 0, None)) == ["edge_" + str(j) + "/0"
                                                    for j in range(num_tokens)]
        for tool in ("", FakeTokenizationAndDependency.default_tool):
            assert list(cdt(tokenization, 0, tool)) == [""] * num_tokens

        for i in range(1, num_tokens+1):
            for tool in (None, "",
                         FakeTokenizationAndDependency.default_tool):
                assert list(cdt(tokenization, i, tool)) == [""] * num_tokens


def test_get_conll_deprel_tags_arbitrary_num_tokens_num_dp_not_found():
    t = "Other great tool"
    # without this memoization, style fails
    style_tool = FakeTokenizationAndDependency.default_tool
    for num_dp in range(1, 4):
        for num_tokens in range(1, 10):
            tokenization = FakeTokenizationAndDependency(num_tokens,
                                                         num_dp=num_dp,
                                                         t=t).tokenization

            # no tool given, with valid index: return indexed parse
            assert list(cdt(tokenization, 0, None)) == ["edge_" + str(i) + "/0"
                                                        for i in range(num_tokens)]
            # tools given, with valid index, but not found
            assert list(cdt(tokenization, 0, style_tool)) == [""] * num_tokens
            assert list(cdt(tokenization, 0, "")) == [""] * num_tokens

            # invalid index, no tool given: return empty
            assert list(cdt(tokenization, -1, None)) == [""] * num_tokens

            for i in range(num_dp):
                # no tool given, valid index: return specified parse
                assert list(cdt(tokenization, i, None)) == [
                    "edge_" + str(j) + "/" + str(i)
                    for j in range(num_tokens)]
                # tool given, valid index, not found: return empty
                for tool in ("",
                             FakeTokenizationAndDependency.default_tool):
                    assert list(cdt(tokenization, i, tool)) == [""] * num_tokens

            # invalid index: return empty
            for tool in (None, "",
                         FakeTokenizationAndDependency.default_tool):
                assert list(cdt(tokenization, num_dp, tool)) == [""] * num_tokens


def test_get_conll_deprel_tags_arbitrary_num_tokens_num_dp_found():
    # sigh... memoize to let style checks pass
    style_tool = FakeTokenizationAndDependency.default_tool
    for num_dp in range(1, 4):
        for num_tokens in range(1, 10):
            tokenization = FakeTokenizationAndDependency(
                num_tokens,
                num_dp=num_dp).tokenization

            # no tool given, with valid index: return indexed parse
            assert list(cdt(tokenization, 0, None)) == ["edge_" + str(i) + "/0"
                                                        for i in range(num_tokens)]
            # tool given, with valid index, not found
            assert list(cdt(tokenization, 0, style_tool)) == [
                "edge_" + str(i) + "/0" for i in range(num_tokens)]
            assert list(cdt(tokenization, 0, "")) == [""] * num_tokens

            # invalid index, no tool given: return empty
            assert list(cdt(tokenization, -1, None)) == [""] * num_tokens

            for i in range(num_dp):
                # tool given, valid index, not found: return empty
                assert list(cdt(tokenization, i, "")) == [""] * num_tokens
                # tool none, valid index, found: return requested
                assert list(cdt(tokenization, i, None)) == [
                    "edge_" + str(j) + "/" + str(i)
                    for j in range(num_tokens)]
                # tool given, valid index, found: return correct
                assert list(cdt(tokenization, i, style_tool)) == [
                    "edge_" + str(j) + "/0" for j in range(num_tokens)]

            # invalid index: return empty
            for tool in (None, ""):
                assert list(cdt(tokenization, num_dp, tool)) == [""] * num_tokens

            assert list(cdt(tokenization, num_dp, style_tool)) == [
                "edge_" + str(j) + "/0" for j in range(num_tokens)]



from __future__ import unicode_literals
from concrete.util import (
    DirectoryBackedCommunicationContainer,
    MemoryBackedCommunicationContainer,
    ZipFileBackedCommunicationContainer,
    RedisHashBackedCommunicationContainer,
)
from concrete.util import create_comm
from concrete.util import write_communication_to_buffer

from concrete.validate import validate_communication

from pytest import raises, fixture
import mock


@fixture
def comm_id_and_buf(request):
    comm_id = 'temp comm'
    return (comm_id, write_communication_to_buffer(create_comm(comm_id)))


def test_directory_backed_comm_container_find_files_recursively():
    directory_path = u'tests/testdata/a'
    cc = DirectoryBackedCommunicationContainer(directory_path)
    assert 3 == len(cc)


def test_directory_backed_comm_container_retrieve():
    directory_path = u'tests/testdata/a'
    cc = DirectoryBackedCommunicationContainer(directory_path)
    assert 3 == len(cc)
    assert u'simple_1' in cc
    for comm_id in cc:
        comm = cc[comm_id]
        assert validate_communication(comm)


def test_memory_backed_comm_container_file_too_large():
    comm_path = u'tests/testdata/simple.tar.gz'
    with raises(Exception):
        MemoryBackedCommunicationContainer(comm_path, max_file_size=500)


def test_memory_backed_comm_container_retrieve():
    comm_path = u'tests/testdata/simple.tar.gz'
    cc = MemoryBackedCommunicationContainer(comm_path)
    assert 3 == len(cc)
    assert u'one' in cc
    for comm_id in cc:
        comm = cc[comm_id]
        assert validate_communication(comm)


def test_zip_file_backed_comm_container_retrieve():
    zipfile_path = u'tests/testdata/simple.zip'
    cc = ZipFileBackedCommunicationContainer(zipfile_path)
    assert 3 == len(cc)
    assert u'simple_1' in cc
    for comm_id in cc:
        comm = cc[comm_id]
        assert validate_communication(comm)


def test_redis_hash_backed_comm_container_len(comm_id_and_buf):
    redis_db = mock.Mock()
    key = mock.sentinel
    (comm_id, comm_buf) = comm_id_and_buf
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    redis_db.hlen.side_effect = [3]
    assert 3 == len(cc)
    redis_db.hlen.assert_called_once_with(key)


def test_redis_hash_backed_comm_container_contains(comm_id_and_buf):
    redis_db = mock.Mock()
    key = mock.sentinel
    (comm_id, comm_buf) = comm_id_and_buf
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    redis_db.hexists.side_effect = [False]
    assert comm_id + 'x' not in cc
    redis_db.hexists.assert_called_once_with(key, comm_id + 'x')


def test_redis_hash_backed_comm_container_not_contains(comm_id_and_buf):
    redis_db = mock.Mock()
    key = mock.sentinel
    (comm_id, comm_buf) = comm_id_and_buf
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    redis_db.hexists.side_effect = [True]
    assert comm_id in cc
    redis_db.hexists.assert_called_once_with(key, comm_id)


def test_redis_hash_backed_comm_container_not_getitem(comm_id_and_buf):
    redis_db = mock.Mock()
    key = mock.sentinel
    (comm_id, comm_buf) = comm_id_and_buf
    cc = RedisHashBackedCommunicationContainer(redis_db, key)

    redis_db.hget.side_effect = [None]
    with raises(KeyError):
        cc[comm_id + 'x']
    redis_db.hget.assert_called_once_with(key, comm_id + 'x')



from __future__ import unicode_literals
from .mem_io import (
    read_communication_from_buffer,
    write_communication_to_buffer
)

from uuid import uuid4

try:
    xrange
except NameError:
    xrange = range


def _identity(o):
    return o


def read_communication_from_redis_key(redis_db, key, add_references=True):
    '''
    Return a serialized communication from a string key.  If block
    is True, poll server until key appears at specified interval
    or until specified timeout (indefinitely if timeout is zero).
    Return None if block is False and key does not exist or if
    block is True and key does not exist after specified timeout.

    Args:
        redis_db:
        key:
        add_references:
    '''
    buf = redis_db.get(key)
    if buf is None:
        return None
    else:
        return read_communication_from_buffer(buf,
                                              add_references=add_references)


class RedisReader(object):
    '''
    Iterable class for reading one or more objects from redis.

    Supported input types are:

    - a set containing zero or more objects
    - a list containing zero or more objects
    - a hash containing zero or more key-object pairs

    For list and set types, the reader can optionally pop (consume) its
    input; for lists only, the reader can moreover block on the input.

    Note that iteration over a set or hash will create a temporary key
    in the redis database to maintain a set of elements scanned so far.

    If pop is False and the key (in the database) is modified during
    iteration, behavior is undefined.  If pop is True, modifications
    during iteration are encouraged.

    Example usage::

        from redis import Redis
        redis_db = Redis(port=12345)
        for obj in RedisReader(redis_db, 'my-set-key'):
            do_something(obj)
    '''

    def __init__(self, redis_db, key, key_type=None, pop=False, block=False,
                 right_to_left=True,
                 block_timeout=0, temp_key_ttl=3600, temp_key_leaf_len=32,
                 cycle_list=False, deserialize_func=None):
        '''
        Create reader for specified key in specified
        redis_db.

        Args:
            redis_db: object of class redis.Redis
            key: name of redis key containing your object(s)
            key_type: 'set', 'list', 'hash', or None; if None, look up
                type in redis (only works if the key exists, so
                probably not suitable for block and/or pop modes)
            pop: boolean, True to remove objects from redis
                as we iterate over them, and False to leave redis
                unaltered
            block: boolean, True to block for data (i.e., wait for
                something to be added to the list if it is empty),
                False to end iteration when there is no more data
            right_to_left: boolean, True to iterate over and index in
                lists from right to left, False to iterate/index
                from left to right
            deserialize_func: function, maps blobs from redis to some more
                friendly representation (e.g., if all your items
                are unicode strings, you might want to specify
                lambda s: s.decode('utf-8')); return blobs
                unchanged if deserialize_func is None
        '''
        self.redis_db = redis_db
        self.key = key

        if key_type is None:
            # try to guess type
            if redis_db.exists(key):
                key_type = redis_db.type(key).decode('utf-8')
            else:
                raise ValueError('can only guess type of key that exists')

        if key_type not in ('set', 'hash', 'list'):
            raise ValueError('unrecognized key type %s' % key_type)

        self.key_type = key_type

        if pop and key_type not in ('set', 'list'):
            raise ValueError('can only pop on set or list')
        if block and key_type not in ('list',):
            raise ValueError('can only block-pop on list')
        if block and not pop:
            raise ValueError('can only block if popping too')
        if cycle_list and block:
            raise ValueError('can only cycle list when not blocking')
        if cycle_list and key_type not in ('list',):
            raise ValueError('can only cycle list on list type')
        if cycle_list and not right_to_left:
            raise ValueError('can only cycle list if processing right-to-left')

        self.pop = pop
        self.block = block
        self.block_timeout = block_timeout

        self.temp_key_ttl = temp_key_ttl
        self.temp_key_leaf_len = temp_key_leaf_len

        self.right_to_left = right_to_left

        self.cycle_list = cycle_list

        self.deserialize_func = (
            deserialize_func
            if (deserialize_func is not None)
            else _identity
        )

    def __iter__(self):
        if self.key_type in ('list', 'set') and self.pop:
            buf = self._pop_buf()
            while buf is not None:
                yield self.deserialize_func(buf)
                buf = self._pop_buf()

        elif self.key_type == 'list' and not self.pop:
            if self.cycle_list:
                buf = self.redis_db.rpoplpush(self.key, self.key)
                i = 0
                while buf is not None and i < self.redis_db.llen(self.key):
                    yield self.deserialize_func(buf)
                    buf = self.redis_db.rpoplpush(self.key, self.key)
                    i += 1
            else:
                for i in xrange(self.redis_db.llen(self.key)):
                    idx = -(i + 1) if self.right_to_left else i
                    buf = self.redis_db.lindex(self.key, idx)
                    yield self.deserialize_func(buf)

        elif self.key_type in ('set', 'hash') and not self.pop:
            if self.key_type == 'set':
                num_objs = self.redis_db.scard(self.key)
                scan = self.redis_db.sscan
                # batch is an iterable of buffers

                def _deserialize(k, batch):
                    return self.deserialize_func(k)
            else:
                num_objs = self.redis_db.hlen(self.key)
                scan = self.redis_db.hscan
                # batch is a dict of key-buffer pairs

                def _deserialize(k, batch):
                    return self.deserialize_func(batch[k])

            get_obj = _deserialize

            temp_key = self._make_temp_key()

            i = 0
            cursor = 0
            while i < num_objs:
                (cursor, batch) = scan(self.key, cursor)
                for k in batch:
                    if i == num_objs:
                        break
                    if self.redis_db.sadd(temp_key, k) > 0:
                        i += 1
                        yield get_obj(k, batch)
                    self.redis_db.expire(temp_key, self.temp_key_ttl)

        else:
            raise Exception('not implemented')

    def __len__(self):
        '''
        Return instantaneous length of dataset.
        '''
        if self.key_type == 'list':
            return self.redis_db.llen(self.key)
        elif self.key_type == 'set':
            return self.redis_db.scard(self.key)
        elif self.key_type == 'hash':
            return self.redis_db.hlen(self.key)
        else:
            raise Exception('not implemented')

    def __getitem__(self, k):
        '''
        Return item at specified list index or hash key;
        never pop or block.

        Args:
            k:
        '''
        if self.key_type in ('list', 'hash'):
            if self.key_type == 'list':
                idx = -(k + 1) if self.right_to_left else k
                buf = self.redis_db.lindex(self.key, idx)
            else:
                buf = self.redis_db.hget(self.key, k)
            return None if buf is None else self.deserialize_func(buf)

        else:
            raise Exception('not implemented')

    def batch(self, n):
        '''
        Return a batch of n objects.  May be faster than
        one-at-a-time iteration, but currently only supported for
        non-popping, non-blocking set configurations.  Support for
        popping, non-blocking sets is planned; see
        http://redis.io/commands/spop .

        Args:
            n:
        '''
        if self.key_type == 'set' and not self.pop and not self.block:
            return [
                self.deserialize_func(buf)
                for buf
                in self.redis_db.srandmember(self.key, n)
            ]

        else:
            raise Exception('not implemented')

    def _pop_buf(self):
        '''
        Pop and return a serialized object, or None if there is
        none (or we blocked and timed out).
        '''
        if self.key_type == 'list':
            if self.block:
                _pop = (
                    self.redis_db.brpop
                    if self.right_to_left
                    else self.redis_db.blpop
                )
                val = _pop(self.key, timeout=self.block_timeout)
                return None if val is None else val[1]
            else:
                _pop = (
                    self.redis_db.rpop
                    if self.right_to_left
                    else self.redis_db.lpop
                )
                return _pop(self.key)

        elif self.key_type == 'set' and not self.block:
            return self.redis_db.spop(self.key)

        else:
            raise Exception('not implemented')

    def _make_temp_key(self):
        '''
        Generate temporary (random, unused) key.  Do not set in redis
        (leave that to the caller).
        '''
        temp_key = None
        while temp_key is None or self.redis_db.exists(temp_key):
            # thanks: http://stackoverflow.com/a/2257449
            temp_key_leaf = str(uuid4())
            temp_key = ':'.join(('temp', self.key, temp_key_leaf))
        return temp_key

    def __str__(self):
        return '%s(%s, %s, %s)' % (type(self).__name__, self.redis_db,
                                   self.key, self.key_type)


class RedisCommunicationReader(RedisReader):
    '''
    Iterable class for reading one or more Communications from redis.
    See RedisReader for further description.

    Example usage::

        from redis import Redis
        redis_db = Redis(port=12345)
        for comm in RedisCommunicationReader(redis_db, 'my-set-key'):
            do_something(comm)
    '''

    def __init__(self, redis_db, key, add_references=True, **kwargs):
        '''
        Create communication reader for specified key in specified
        redis_db.

        Args:
            redis_db: object of class redis.Redis
            key: name of redis key containing your communication(s)
            add_references: boolean, True to fill in members in the
                communication according to UUID relationships (see
                concrete.util.add_references), False to return
                communication as-is (note: you may need this False
                if you are dealing with incomplete communications)

        All other keyword arguments are passed through to RedisReader.
        '''

        if 'deserialize_func' in kwargs:
            raise ValueError('RedisCommunicationReader does not allow custom '
                             'deserialize_func')
        self.add_references = add_references
        super(RedisCommunicationReader, self).__init__(
            redis_db, key, deserialize_func=self._load_from_buffer, **kwargs
        )

    def _load_from_buffer(self, buf):
        return read_communication_from_buffer(
            buf, add_references=self.add_references)


def write_communication_to_redis_key(redis_db, key, comm):
    '''
    Serialize communication and store result in redis key.
    '''
    redis_db.set(key, write_communication_to_buffer(comm))


class RedisWriter(object):
    '''
    Class for writing one or more objects to redis.

    Supported input types are:

    - a set of objects
    - a list of objects
    - a hash of key-object pairs

    Example usage:

        from redis import Redis
        redis_db = Redis(port=12345)
        w = RedisWriter(redis_db, 'my-set-key')
        w.write(obj)
    '''

    def __init__(self, redis_db, key, key_type=None, right_to_left=True,
                 serialize_func=None, hash_key_func=None):
        '''
        Create object writer for specified key in specified
        redis_db.

        Args:
            redis_db: object of class redis.Redis
            key: name of redis key containing your object(s)
            key_type: 'set', 'list', 'hash', or None; if None, look up
                type in redis (only works if the key exists)
            right_to_left: boolean, True to write elements to the left
                end of lists, False to write to the right end
            serialize_func: function, maps objects to blobs before
                sending to Redis (e.g., if everything you write
                will be a unicode string, you might want to use
                lambda u: u.encode('utf-8')); pass objects to
                Redis unchanged if serialize_func is None
            hash_key_func: function, maps objects to keys when key_type
                is hash (None: use Python's hash function)
        '''
        self.redis_db = redis_db
        self.key = key

        if key_type is None:
            # try to guess type
            if redis_db.exists(key):
                key_type = redis_db.type(key).decode('utf-8')
            else:
                raise ValueError('can only guess type of key that exists')

        if key_type not in ('set', 'hash', 'list'):
            raise ValueError('unrecognized key type %s' % key_type)

        self.key_type = key_type
        self.right_to_left = right_to_left

        self.serialize_func = (
            serialize_func
            if (serialize_func is not None)
            else _identity
        )
        self.hash_key_func = (
            hash_key_func
            if (hash_key_func is not None)
            else hash
        )

    def clear(self):
        self.redis_db.delete(self.key)

    def write(self, obj):
        buf = self.serialize_func(obj)

        if self.key_type == 'list':
            _push = (
                self.redis_db.lpush
                if self.right_to_left
                else self.redis_db.rpush
            )
            return _push(self.key, buf)

        elif self.key_type == 'set':
            return self.redis_db.sadd(self.key, buf)

        elif self.key_type == 'hash':
            return self.redis_db.hset(self.key, self.hash_key_func(obj), buf)

        else:
            raise Exception('not implemented')

    def __str__(self):
        return '%s(%s, %s, %s)' % (type(self).__name__, self.redis_db,
                                   self.key, self.key_type)


class RedisCommunicationWriter(RedisWriter):
    '''
    Class for writing one or more Communications to redis.
    See RedisWriter for further description.

    Example usage:

        from redis import Redis
        redis_db = Redis(port=12345)
        w = RedisCommunicationWriter(redis_db, 'my-set-key')
        w.write(comm)
    '''

    def __init__(self, redis_db, key, uuid_hash_key=False, **kwargs):
        '''
        Create communication writer for specified key in specified
        redis_db.

        Args:
            redis_db: object of class redis.Redis
            key: name of redis key containing your communication(s)
            uuid_hash_key: boolean, True to use the UUID as the hash key
                 for a communication, False to use the id
        '''
        if 'serialize_func' in kwargs:
            raise ValueError('RedisCommunicationWriter does not allow custom '
                             'serialize_func')
        if 'hash_key_func' in kwargs:
            raise ValueError('RedisCommunicationWriter does not allow custom '
                             'hash_key_func')
        self.uuid_hash_key = uuid_hash_key
        super(RedisCommunicationWriter, self).__init__(
            redis_db, key, serialize_func=self._write_to_buffer,
            hash_key_func=self._to_key, **kwargs
        )

    def _write_to_buffer(self, comm):
        return write_communication_to_buffer(comm)

    def _to_key(self, comm):
        return (comm.uuid.uuidString if self.uuid_hash_key else comm.id)

    def __str__(self):
        return '%s(%s, %s, %s)' % (type(self).__name__, self.redis_db,
                                   self.key, self.key_type)



from __future__ import unicode_literals


def lun(l):
    '''
    If l is None return an empty list, else return l.
    Simplifies iteration over list fields that might be unset.
    '''
    return list() if (l is None) else l


def dun(d):
    '''
    If l is None return an empty dict, else return l.
    Simplifies iteration over dict fields that might be unset.
    '''
    return dict() if (d is None) else d


def sun(s):
    '''
    If l is None return an empty set, else return l.
    Simplifies iteration over set fields that might be unset.
    '''
    return set() if (s is None) else s
